#include <assert.h>

#include <assimp/cimport.h>
#include <assimp/material.h>
#include <assimp/mesh.h>
#include <assimp/postprocess.h>
#include <assimp/scene.h>

#include <string>

#include <gmath/gmath.h>

#include "mesh.h"
#include "object.h"
#include "scene.h"
#include "texture.h"

#include "opengl/mesh-gl.h"
#include "vulkan/mesh-vk.h"

#include "opengl/texture-gl.h"
#include "vulkan/texture-vk.h"

extern bool use_vulkan;
static Mesh *load_mesh(const aiScene *scene, unsigned int index);
static Material *load_material(const aiScene *ascene, Scene *scene, unsigned int index);

Scene::Scene() {}

Scene::~Scene()
{
	for(size_t i=0; i<meshes.size(); ++i)
		delete meshes[i];
	meshes.clear();

	for(size_t i=0; i<materials.size(); ++i)
		delete materials[i];
	materials.clear();

	for(size_t i=0; i<textures.size(); ++i)
		delete textures[i];
	textures.clear();

	for(size_t i=0; i<objects.size(); ++i)
		delete objects[i];
	objects.clear();
}

bool Scene::load(const char *fname)
{
	/* loading scene */
	unsigned int ai_flags = aiProcessPreset_TargetRealtime_Quality;
	const aiScene *scene = aiImportFile(fname, ai_flags);
	if(!scene) {
		fprintf(stderr, "Failed to import scene: %s\n", fname);
		return false;
	}

	/* loading scene meshes */
	for(unsigned int i=0; i<scene->mNumMeshes; ++i) {
		Mesh *mesh = load_mesh(scene, i);
		if(!mesh) {
			fprintf(stderr, "Failed to load mesh no: %d.\n", i);
			return false;
		}
		meshes.push_back(mesh);
	}

	/* loading materials */
	for(unsigned int i=0; i<scene->mNumMaterials; ++i) {
		Material *material = load_material(scene, this, i);
		if(!material) {
			fprintf(stderr, "Failed to load material no: %d", i);
			return false;
		}
		materials.push_back(material);
	}

	/* create objects */
	for(unsigned int i=0; i<scene->mNumMeshes; ++i) {
		Object *object = new Object();
		object->mesh = meshes[i];
		int idx = meshes[i]->mat_idx;
		object->material = materials[idx];
		objects.push_back(object);
	}

	aiReleaseImport(scene);
	return true;
}

static Mesh *load_mesh(const aiScene *scene, unsigned int index)
{
	/* load mesh with index from scene using assimp */
	aiMesh *amesh = scene->mMeshes[index];
	if(!amesh) {
		fprintf(stderr, "Failed to load assimp mesh no: %d.\n", index);
		return 0;
	}

	Mesh *mesh;
	if(use_vulkan) {
		mesh = new MeshVK;
	}
	else {
		mesh = new MeshGL;
	}

	mesh->name = std::string(amesh->mName.data);
	mesh->which_mask = 0;

	for(unsigned int i=0; i<amesh->mNumVertices; ++i) {
		/* vertices */
		if(amesh->HasPositions()) {
			mesh->which_mask |= MESH_VERTEX;
			Vec3 vertex = Vec3(amesh->mVertices[i].x, amesh->mVertices[i].y,
			                   amesh->mVertices[i].z);

			mesh->vertices.push_back(vertex);
		}
		else {
			fprintf(stderr, "Mesh has no geometry!\n");
			delete mesh;
			return 0;
		}

		/* normals */
		if(amesh->HasNormals()) {
			mesh->which_mask |= MESH_NORMAL;
			Vec3 normal = Vec3(amesh->mNormals[i].x, amesh->mNormals[i].y,
			                   amesh->mNormals[i].z);
			mesh->normals.push_back(normal);
		}
		else {
			fprintf(stderr, "Mesh has no normals!\n");
			delete mesh;
			return 0;
		}

		/* texture coordinates */
		if(amesh->mTextureCoords[0]) {
			mesh->which_mask |= MESH_TEXTURE;
			Vec2 tex_coord =
			    Vec2(amesh->mTextureCoords[0][i].x, amesh->mTextureCoords[0][i].y);
			mesh->tex_coords.push_back(tex_coord);
		}
		else {
			mesh->tex_coords.push_back(Vec2(0, 0));
		}

		/* tangents */
		if(amesh->mTangents) {
			mesh->which_mask |= MESH_TANGENT;
			Vec3 tangent = Vec3(amesh->mTangents[i].x, amesh->mTangents[i].y,
			                    amesh->mTangents[i].z);
			mesh->tangents.push_back(tangent);
		}
		else {
			mesh->tangents.push_back(Vec3(1, 0, 0));
		}
	}
	/* indices (called faces in assimp) */
	if(amesh->HasFaces()) {
		mesh->which_mask |= MESH_INDEX;
		for(unsigned int i=0; i<amesh->mNumFaces; ++i) {
			mesh->indices.push_back(amesh->mFaces[i].mIndices[0]);
			mesh->indices.push_back(amesh->mFaces[i].mIndices[1]);
			mesh->indices.push_back(amesh->mFaces[i].mIndices[2]);
		}
	}
	else
		fprintf(stderr, "No faces found.\n");

	mesh->mat_idx = amesh->mMaterialIndex;
	printf("material idx:%u", mesh->mat_idx);
	return mesh;
}

static Material *load_material(const aiScene *ascene, Scene *scene, unsigned int index)
{
	aiMaterial *amat = ascene->mMaterials[index];
	if(!amat) {
		fprintf(stderr, "Failed to load material no: %d.\n", index);
		return 0;
	}

	Material *mat = new Material;
	mat->dtex = 0;
	mat->shininess = 40;

	// aiString name;
	// amat->Get(AI_MATKEY_NAME, name);
	// mat->name = std::string(name.data);

	aiColor4D color;
	aiGetMaterialColor(amat, AI_MATKEY_COLOR_DIFFUSE, &color);
	mat->diffuse = Vec3(color.r, color.g, color.b);

	aiGetMaterialColor(amat, AI_MATKEY_COLOR_SPECULAR, &color);
	float spstr;
	aiGetMaterialFloat(amat, AI_MATKEY_SHININESS_STRENGTH, &spstr);
	mat->specular = spstr * Vec3(color.r, color.g, color.b);

	aiGetMaterialFloat(amat, AI_MATKEY_SHININESS, &mat->shininess);

	aiString tex_name;
	if(aiGetMaterialString(amat, AI_MATKEY_TEXTURE_DIFFUSE(0), &tex_name) == aiReturn_SUCCESS) {
		/* different scene objects might use the same texture, we shouldn't store it twice*/
		//mat->dtex = scene->find_texture(tex_name.data);
		if(!mat->dtex) {
			if(use_vulkan) {
				mat->dtex = new TextureVK;
			}
			else {
				mat->dtex = new TextureGL;
			}
			if(!mat->dtex->load(tex_name.data)) {
				fprintf(stderr, "Failed to load texture data: %s.\n", tex_name.data);
				delete mat->dtex;
				mat->dtex = 0;
			}
			scene->textures.push_back(mat->dtex);
		}
	}

	return mat;
}
